from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Callable, Literal

from embar.custom_types import Type

type OnDelete = (
    Literal["no action"] | Literal["restrict"] | Literal["set null"] | Literal["set default"] | Literal["cascade"]
)


@dataclass
class ColumnInfo:
    """
    `ColumnInfo` is the type that ultimately holds all the db column info.

    It knows nothing about the python field: its name, what type it should deserialize to etc.
    """

    # _table_name is callable as generally the `Table` won't yet have a name
    # at the time the Column is created.
    _table_name: Callable[[], str]

    name: str
    col_type: str
    py_type: Type
    primary: bool
    not_null: bool
    default: Any | None = None

    ref: ColumnInfo | None = None
    on_delete: OnDelete | None = None

    args: str | None = None

    @property
    def table_name(self) -> str:
        return self._table_name()

    def fqn(self) -> str:
        """
        Return the Fully Qualified Name (table and column both in quotes).

        ```python
        from embar.column.base import ColumnInfo
        col = ColumnInfo(
           _table_name=lambda: "foo", name="bar", col_type="TEXT", py_type=str, primary=False, not_null=False
        )
        fqn = col.fqn()
        assert fqn == '"foo"."bar"'
        ```
        """
        return f'"{self._table_name()}"."{self.name}"'

    def ddl(self) -> str:
        """
        Generate the DDL just for this column.

        Used by the [`Table.ddl()`][embar.table.Table.ddl] method to generate the full DDL.

        ```python
        from embar.column.base import ColumnInfo
        col = ColumnInfo(
           _table_name=lambda: "foo", name="bar", col_type="TEXT", py_type=str, primary=True, not_null=True
        )
        ddl = col.ddl()
        assert ddl == '"bar" TEXT NOT NULL PRIMARY KEY'
        ```
        """
        args = self.args if self.args is not None else ""
        default = f"DEFAULT '{self.default}'" if self.default is not None else ""
        nullable = "NOT NULL" if self.not_null else ""
        primary = "PRIMARY KEY" if self.primary else ""
        reference = f'REFERENCES "{self.ref.table_name}"("{self.ref.name}")' if self.ref is not None else ""
        on_delete = f"ON DELETE {self.on_delete}" if self.on_delete is not None else ""
        text = f'"{self.name}" {self.col_type}{args} {default} {nullable} {primary} {reference} {on_delete}'
        clean = " ".join(text.split()).strip()
        return clean


class ColumnBase:
    """
    Base class for all [`Column`][embar.column.common.Column] classes.

    Mostly here to avoid circular dependencies with modules that need to know about the fields below.
    """

    info: ColumnInfo  # pyright:ignore[reportUninitializedInstanceVariable]

    # These must always be assigned by children, type-checker won't catch it
    _sql_type: str  # pyright:ignore[reportUninitializedInstanceVariable]
    _py_type: Type  # pyright:ignore[reportUninitializedInstanceVariable]


class EnumBase(ABC):
    name: str

    @classmethod
    @abstractmethod
    def ddl(cls) -> str: ...



from typing import Annotated

from pydantic import BaseModel

from embar.db.pg import PgDb
from embar.db.sqlite import SqliteDb
from embar.query.where import Eq

from ..schemas.schema import Message, MessageUpdate


def test_update_row(db_loaded: SqliteDb | PgDb):
    db = db_loaded

    new_content = "new content"
    # fmt: off
    (
        db.update(Message)
        .set(MessageUpdate(content=new_content))
        .where(Eq(Message.id, 1))
        .run()
    )
    # fmt: on

    class MessageSel(BaseModel):
        content: Annotated[str, Message.content]

    res = (
        db.select(MessageSel)
        .fromm(Message)
        .where(
            Eq(Message.id, 1),
        )
        .limit(1)
        .run()
    )

    assert len(res) == 1
    got = res[0]
    assert got.content == new_content



"""Base classes for database clients."""

from abc import ABC, abstractmethod
from typing import Any, Literal

from embar.custom_types import Undefined
from embar.query.query import Query

DbType = Literal["sqlite"] | Literal["postgres"]


class AllDbBase:
    """
    Base class (not an ABC, but could be) for all Db clients.
    """

    db_type: DbType = Undefined


class DbBase(ABC, AllDbBase):
    """
    Base class for _sync_ Db clients.
    """

    @abstractmethod
    def execute(self, query: Query):
        """
        Execute a query without returning results.
        """
        ...

    @abstractmethod
    def executemany(self, query: Query):
        """
        Execute a query with multiple parameter sets.
        """
        ...

    @abstractmethod
    def fetch(self, query: Query) -> list[dict[str, Any]]:
        """
        Execute a query and return results as a list of dicts.
        """
        ...

    @abstractmethod
    def truncate(self, schema: str | None = None) -> None:
        """
        Truncate all tables in the schema.
        """
        ...


class AsyncDbBase(ABC, AllDbBase):
    """
    Base class for async Db clients.
    """

    @abstractmethod
    async def execute(self, query: Query):
        """
        Execute a query without returning results.
        """
        ...

    @abstractmethod
    async def executemany(self, query: Query):
        """
        Execute a query with multiple parameter sets.
        """
        ...

    @abstractmethod
    async def fetch(self, query: Query) -> list[dict[str, Any]]:
        """
        Execute a query and return results as a list of dicts.
        """
        ...

    @abstractmethod
    async def truncate(self, schema: str | None = None) -> None:
        """
        Truncate all tables in the schema.
        """
        ...



"""Join clauses for queries."""

from abc import ABC, abstractmethod
from typing import override

from embar.query.query import Query
from embar.query.where import GetCount, WhereClause
from embar.table import Table


class JoinClause(ABC):
    """
    Base class for all join clauses.
    """

    @abstractmethod
    def get(self, get_count: GetCount) -> Query:
        """
        Generate the SQL for this join clause.
        """
        ...


class LeftJoin(JoinClause):
    """
    LEFT JOIN clause.
    """

    table: type[Table]
    on: WhereClause

    def __init__(self, table: type[Table], on: WhereClause):
        """
        Create a new LeftJoin instance.
        """
        self.table = table
        self.on = on

    @override
    def get(self, get_count: GetCount) -> Query:
        """
        Generate the LEFT JOIN SQL.
        """
        on = self.on.sql(get_count)

        sql = f"LEFT JOIN {self.table.fqn()} ON {on.sql}"
        return Query(sql=sql, params=on.params)


class RightJoin(JoinClause):
    """
    RIGHT JOIN clause.
    """

    table: type[Table]
    on: WhereClause

    def __init__(self, table: type[Table], on: WhereClause):
        """
        Create a new RightJoin instance.
        """
        self.table = table
        self.on = on

    @override
    def get(self, get_count: GetCount) -> Query:
        """
        Generate the RIGHT JOIN SQL.
        """
        on = self.on.sql(get_count)

        sql = f"RIGHT JOIN {self.table.fqn()} ON {on.sql}"
        return Query(sql=sql, params=on.params)


class InnerJoin(JoinClause):
    """
    INNER JOIN clause.
    """

    table: type[Table]
    on: WhereClause

    def __init__(self, table: type[Table], on: WhereClause):
        """
        Create a new InnerJoin instance.
        """
        self.table = table
        self.on = on

    @override
    def get(self, get_count: GetCount) -> Query:
        """
        Generate the INNER JOIN SQL.
        """
        on = self.on.sql(get_count)

        sql = f"INNER JOIN {self.table.fqn()} ON {on.sql}"
        return Query(sql=sql, params=on.params)


class FullJoin(JoinClause):
    """
    FULL OUTER JOIN clause.
    """

    table: type[Table]
    on: WhereClause

    def __init__(self, table: type[Table], on: WhereClause):
        """
        Create a new FullJoin instance.
        """
        self.table = table
        self.on = on

    @override
    def get(self, get_count: GetCount) -> Query:
        """
        Generate the FULL OUTER JOIN SQL.
        """
        on = self.on.sql(get_count)

        sql = f"FULL OUTER JOIN {self.table.fqn()} ON {on.sql}"
        return Query(sql=sql, params=on.params)


class CrossJoin(JoinClause):
    """
    CROSS JOIN clause.
    """

    table: type[Table]

    def __init__(self, table: type[Table]):
        """
        Create a new CrossJoin instance.
        """
        self.table = table

    @override
    def get(self, get_count: GetCount) -> Query:
        """
        Generate the CROSS JOIN SQL.
        """
        sql = f"CROSS JOIN {self.table.fqn()}"
        return Query(sql=sql)
